#include <string.h>
#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/disk.h"

#include "guiutils.h"
#include "toolbar.h"
#include "dspview.h"

#include "cfg.h"
#include "win.h"
#include "wincb.h"
#include "windnd.h"
#include "winopcb.h"
#include "windpy.h"
#include "core.h"
#include "cfglist.h"
#include "config.h"

#include "images/icon_yplayer_48x48.xpm"

#include "images/icon_open_20x20.xpm"
#include "images/icon_close_20x20.xpm"
#include "images/icon_yserver_location_20x20.xpm"
#include "images/icon_reload_20x20.xpm"
#include "images/icon_tool_bar_20x20.xpm"
#include "images/icon_status_bar_20x20.xpm"
#include "images/icon_yplayer_20x20.xpm"
#include "images/icon_ymixer_20x20.xpm"
#include "images/icon_about_20x20.xpm"

#include "images/icon_yplayer_eject_20x20.xpm"
/* #include "images/icon_yplayer_repeat_20x20.xpm" */
#include "images/icon_yplayer_pause_20x20.xpm"
#include "images/icon_yplayer_rewind_20x20.xpm"
#include "images/icon_yplayer_play_20x20.xpm"
#include "images/icon_yplayer_stop_20x20.xpm"          

#include "images/icon_yplayer_eject_14x14.xpm"
/* #include "images/icon_yplayer_repeat_14x14.xpm" */
#include "images/icon_yplayer_pause_14x14.xpm"
#include "images/icon_yplayer_rewind_14x14.xpm"
#include "images/icon_yplayer_play_14x14.xpm"
#include "images/icon_yplayer_stop_14x14.xpm"

#include "images/icon_speaker_left_20x20.xpm"
#include "images/icon_speaker_right_20x20.xpm"


/* Connection */
void WinConnectionOpen(win_struct *win, const gchar *address);
void WinConnectionClose(win_struct *win);

/* Play IO */
gboolean WinIsPlaying(win_struct *win);
gulong WinGetPosition(win_struct *win);
void WinSetPosition(win_struct *win, gulong pos);
gfloat *WinGetVolume(win_struct *win);
void WinSetVolume(win_struct *win, const gfloat *vol);

/* Play Control */
void WinOpen(win_struct *win, const gchar *path);
void WinPlay(win_struct *win);
void WinPause(win_struct *win);
void WinStop(win_struct *win);
void WinRewind(win_struct *win);

/* Management */
void WinManage(win_struct *win);

/* DSP View */
void WinDSPViewSetDisplayStyle(
	win_struct *win, dspview_display_style display_style
);
void WinDSPViewDisplayStyleNext(win_struct *win);
void WinDSPViewDisplayStylePrev(win_struct *win);

/* OPIDs */
win_opid_struct *WinOPIDNew(
	win_opid id, win_struct *win
);
void WinOPIDDelete(win_opid_struct *opid);
GList *WinOPIDListNew(win_struct *win);
void WinOPIDListDelete(GList *glist);
win_opid_struct *WinOPIDListFind(GList *glist, win_opid id);

/* Win */
static void WinBuildMenuBar(win_struct *win);
static void WinBuildToolBar(win_struct *win);
static void WinBuildLists(win_struct *win);
static void WinBuildStatusBar(win_struct *win);
static void WinBuildListsMenu(win_struct *win);

win_struct *WinNew(gpointer core_ptr);
void WinUpdate(win_struct *win);
void WinSetBusy(win_struct *win, gboolean busy);
gboolean WinToolBarIsShown(win_struct *win);
void WinToolBarSetShow(win_struct *win, gboolean show);
void WinToolBarUpdateItems(win_struct *win);
gboolean WinDisplayIsShown(win_struct *win);
void WinDisplaySetShow(win_struct *win, gboolean show);
gboolean WinStatusBarIsShown(win_struct *win);
void WinStatusBarSetShow(win_struct *win, gboolean show);
void WinStatusProgress(
	win_struct *win, gfloat val, gboolean allow_gtk_iteration
);
void WinStatusMessage(
	win_struct *win, const gchar *mesg, gboolean allow_gtk_iteration
);
gboolean WinIsMapped(win_struct *win);          
void WinMap(win_struct *win);         
void WinUnmap(win_struct *win);
void WinDelete(win_struct *win);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


#define WIN_DSPVIEW_STYLES	{				\
 { DSPVIEW_DISPLAY_STYLE_WAVE, DSPVIEW_MERGE_MODE_ADD,		\
   0xffffb0c0, 0.0f						\
 },								\
 { DSPVIEW_DISPLAY_STYLE_SGRAPH, DSPVIEW_MERGE_MODE_ADD,	\
   0xffffb0c0, 0.0f						\
 },								\
 { DSPVIEW_DISPLAY_STYLE_ANALYZER, DSPVIEW_MERGE_MODE_ADD,	\
   0xffff90b0, 1.0f						\
 },								\
 { DSPVIEW_DISPLAY_STYLE_LIGHTENING, DSPVIEW_MERGE_MODE_ADD,	\
   0xffff80c0, 0.25f						\
 },								\
 { DSPVIEW_DISPLAY_STYLE_PLASMA, DSPVIEW_MERGE_MODE_ADD,	\
   0xffffb0c0, 0.35f						\
 },								\
 { DSPVIEW_DISPLAY_STYLE_NOVA, DSPVIEW_MERGE_MODE_ADD,		\
   0xffffb0c0, 0.32f						\
 }								\
}
typedef struct {
	dspview_display_style	display_style;
	dspview_merge_mode	merge_mode;
	guint32			color;		/* 0xAABBGGRR */
	gfloat			fade_coeff;
} win_dspview_style_struct;


#define WIN_WIDTH			-1
#define WIN_HEIGHT			-1
#define WIN_STATUS_BAR_HEIGHT		26

#define WIN_DISPLAY_WIDTH		210
#define WIN_DISPLAY_HEIGHT		50

#define WIN_DSPVIEW_WIDTH		100
#define WIN_DSPVIEW_HEIGHT		50

#define WIN_BTN_WIDTH			60
#define WIN_BTN_HEIGHT			20


/*
 *	Opens the connection to the Y Server.
 */
void WinConnectionOpen(win_struct *win, const gchar *address)
{
	YConnection *con;

	if(win == NULL)
	    return;

	/* Close existing connection (if any) */
	WinConnectionClose(win);

	WinStatusMessage(win, "Connecting to Y Server...", TRUE);

	/* Connect to Y Server */
	win->con = con = YOpenConnection(
	    NULL, address
	);

	/* Connected to Y Server? */
	if(con != NULL)
	{
	    gchar *buf;

	    /* Update Y Server address */
	    if(win->recorder != address)
	    {
		g_free(win->recorder);
		win->recorder = STRDUP(address);
	    }

	    /* Get YSHM sound buffer pointer */
	    win->yshm_buf = (guint8 *)YSHMOpenSound(
		con, &win->yshm_buf_len, &win->yshm_id
	    );

	    if(!STRISEMPTY(address))
		buf = g_strdup_printf(
		    "Connected to Y Server \"%s\"",
		    address
		);
	    else
		buf = STRDUP("Connected to Y Server");
	    WinStatusMessage(win, buf, TRUE);
	    g_free(buf);
	}
	else
	{
	    WinStatusMessage(win, "Connect failed", TRUE);
	}

	WinDpyQueueDraw(win);
	WinUpdate(win);
}

/*
 *	Closes the connection to the Y Server.
 */
void WinConnectionClose(win_struct *win)
{
	if(win == NULL)
	    return;

	/* Stop sound object from playing (if any) */
	WinStop(win);

	WinStatusMessage(win, "Disconnecting from Y Server...", TRUE);

	/* Close YSHM sound buffer. */
	YSHMCloseSound(win->con, win->yshm_buf);
	win->yshm_buf = NULL;
	win->yshm_buf_len = 0;
	win->yshm_id = -1;

	/* Disconnection to Y server */
	YCloseConnection(win->con, False);
	win->con = NULL;

	WinStatusMessage(win, "Disconnected from Y Server", TRUE);

	WinDpyQueueDraw(win);
	WinUpdate(win);
}


/*
 *	Checks if the sound object is playing.
 */
gboolean WinIsPlaying(win_struct *win)
{
	return((win != NULL) ? (win->yid != YIDNULL) : FALSE);
}

/*
 *	Gets the current position.
 */
gulong WinGetPosition(win_struct *win)
{
	GtkRange *range = (win != NULL) ?
	    GTK_RANGE(win->position_scale) : NULL;
	if(range == NULL)
	    return(0l);

	return((gulong)GTK_ADJUSTMENT_GET_VALUE(range->adjustment));
}

/*
 *	Sets the position.
 */
void WinSetPosition(win_struct *win, gulong pos)
{
	gulong len;
	GtkAdjustment *adj;
	GtkRange *range = (win != NULL) ?
	    GTK_RANGE(win->position_scale) : NULL;
	if(range == NULL)
	    return;

	adj = range->adjustment;
	len = (gulong)adj->upper;
	if(pos > len)
	    pos = len;
	if(pos < 0l)
	    pos = 0l;

	gtk_adjustment_set_value(adj, (gfloat)pos);

	WinDpyQueueDraw(win);
	WinUpdate(win);
}

/*
 *	Gets the volume as 4 gfloats (front left, front right, back
 *	left, and back right).
 */
gfloat *WinGetVolume(win_struct *win)
{
	static gfloat vol[4];
	GtkRange *range;

	memset(vol, 0x00, 4 * sizeof(gfloat));

	if(win == NULL)
	    return(vol);

	range = GTK_RANGE(win->volume_fleft_scale);
	vol[0] = 1.0f - GTK_ADJUSTMENT_GET_VALUE(range->adjustment);
	range = GTK_RANGE(win->volume_fright_scale);
	vol[1] = 1.0f - GTK_ADJUSTMENT_GET_VALUE(range->adjustment);
	range = GTK_RANGE(win->volume_bleft_scale);
	vol[2] = 1.0f - GTK_ADJUSTMENT_GET_VALUE(range->adjustment);
	range = GTK_RANGE(win->volume_bright_scale);
	vol[3] = 1.0f - GTK_ADJUSTMENT_GET_VALUE(range->adjustment);

	return(vol);
}

/*
 *	Sets the volume as 4 gfloats (front left, front right, back     
 *	left, and back right).
 */
void WinSetVolume(win_struct *win, const gfloat *vol)
{
	GtkRange *range;

	if((win == NULL) || (vol == NULL))
	    return;      

	range = GTK_RANGE(win->volume_fleft_scale);
	gtk_adjustment_set_value(range->adjustment, 1.0f - vol[0]);
	range = GTK_RANGE(win->volume_fright_scale);
	gtk_adjustment_set_value(range->adjustment, 1.0f - vol[1]);
	range = GTK_RANGE(win->volume_bleft_scale);
	gtk_adjustment_set_value(range->adjustment, 1.0f - vol[2]);
	range = GTK_RANGE(win->volume_bright_scale);
	gtk_adjustment_set_value(range->adjustment, 1.0f - vol[3]);

	WinDpyQueueDraw(win);
	WinUpdate(win);      
}


/*
 *	Open.
 */
void WinOpen(win_struct *win, const gchar *path)
{
	core_struct *core;

	if(win == NULL)
	    return;

	core = CORE(win->core);

	WinStop(win);

	g_free(win->cur_path);
	win->cur_path = NULL;

	/* Sound object path specified and connected to Y Server? */
	if(!STRISEMPTY(path) && (win->con != NULL))
	{
	    YConnection *con = win->con;
	    YEventSoundObjectAttributes v;

	    /* Set new sound object path */
	    win->cur_path = STRDUP(path);

	    WinStatusMessage(win, "Opening Sound Object...", TRUE);

	    /* Get sound object attributes */
	    if(!YGetSoundObjectAttributes(con, path, &v))
	    {
		const gulong nsamples = (gulong)v.length;
		gchar *buf;
		const gchar *name;
		GtkRange *range = GTK_RANGE(win->position_scale);
		GtkAdjustment *adj = range->adjustment;

		win->freeze_count++;

		/* Get name portion of the Sound Object */
		name = strrchr(path, DIR_DELIMINATOR);
		if(name != NULL)
		    name++;
		else
		    name = path;

		/* Handle by Sound Object format */
		switch(v.format)
		{
		  case SndObjTypeDSP:
		    adj->value = 0.0f;
		    adj->lower = 0.0f;
		    adj->upper = (gfloat)nsamples;
		    adj->step_increment = (gfloat)MAX(nsamples / 100, 1);
 		    adj->page_increment = (gfloat)MAX(nsamples / 10, 1);
 		    adj->page_size = 0.0f;
		    gtk_adjustment_changed(adj);
		    gtk_adjustment_value_changed(adj);
		    break;

		  case SndObjTypeMIDI:
		    adj->value = 0.0f;
		    adj->lower = 0.0f;
		    adj->upper = 0.0f;
		    adj->step_increment = (gfloat)MAX(nsamples / 100, 1);
		    adj->page_increment = (gfloat)MAX(nsamples / 10, 1);
		    adj->page_size = 0.0f;
		    gtk_adjustment_changed(adj);
		    gtk_adjustment_value_changed(adj);
		    break;
		}
		win->sample_rate = v.sample_rate;
		win->sample_size = v.sample_size;
		win->channels = v.channels;

		/* Volume */
		range = GTK_RANGE(win->volume_fleft_scale);
		gtk_adjustment_set_value(range->adjustment, 0.0f);
		range = GTK_RANGE(win->volume_fright_scale);
		gtk_adjustment_set_value(range->adjustment, 0.0f);
		range = GTK_RANGE(win->volume_bleft_scale);
		gtk_adjustment_set_value(range->adjustment, 0.0f);
		range = GTK_RANGE(win->volume_bright_scale);
		gtk_adjustment_set_value(range->adjustment, 0.0f);


		win->freeze_count--;

		/* Update title & status message */
		buf = g_strdup_printf(
		    "%s: %s",
		    (core->title != NULL) ? core->title : PROG_NAME,
		    name
		);
		gtk_window_set_title(GTK_WINDOW(win->toplevel), buf);
		g_free(buf);

		buf = g_strdup_printf(
		    "Opened Sound Object \"%s\"",
		    name
		);
		WinStatusMessage(win, buf, TRUE);
		g_free(buf);
	    }
	    else
	    {
		WinStatusMessage(win, "Opening Sound Object failed", TRUE);
 	    }
	}

	WinDpyQueueDraw(win);
	WinUpdate(win);
}

/*
 *	Play.
 */
void WinPlay(win_struct *win)
{
	const gfloat *vol;
	const gchar *path;
	YConnection *con;
	YEventSoundPlay v;

	if(win == NULL)
	    return;

	WinStatusMessage(win, "Playing...", TRUE);

	/* Destroy the playing Sound Object (if any) */
	YDestroyPlaySoundObject(win->con, win->yid);
	win->yid = YIDNULL;   

	/* Get loaded Sound Object */
	con = win->con;
	path = win->cur_path;
	if(STRISEMPTY(path) || (con == NULL))
	{
	    WinStatusMessage(win, "Play failed", TRUE);
	    return;
	}

	/* Automatically change the Audio mode as needed? */
	if(win->auto_change_audio_mode &&
	   (win->sample_rate > 0) && (win->sample_size > 0)
	)
	{
	    YAudioModeValuesStruct v;

	    /* Find a Y Audio mode with suitable Audio values */
	    if(YMatchAudioModeValues(
		con, &v,
		win->sample_rate,
		win->sample_size,
		win->channels,
		0
	    ))
	    {
		/* Found a Y Audio mode, now switch to it */
		YChangeAudioModePreset(con, v.name);
	    }
	    else
	    {
		/* Unable to find a suitable Y Audio mode, so set
		 * exact Audio values
		 */
		const gint fragment_size = 1024;

		/* Calculate cycle in microseconds */
		gulong cycle_us = (gulong)YCalculateCycle(
		    con,
		    win->sample_rate,
		    win->channels,
		    win->sample_size,
		    fragment_size
		);

		/* Set Audio values */
		YSetAudioModeValues(
		    con,
		    win->sample_size,
		    win->channels,
		    win->sample_rate,
		    0,
		    True,
		    2,
		    fragment_size
		);

		/* Set cycle */
		YSetCycle(con, (long)cycle_us);
	    }
	}

	/* Set up play values */
	v.flags = YPlayValuesFlagPosition |
	    YPlayValuesFlagTotalRepeats |
	    YPlayValuesFlagVolume |
	    YPlayValuesFlagSampleRate;
	v.position = (YDataPosition)WinGetPosition(win);
	v.total_repeats = win->repeating ? -1 : 1;
	vol = WinGetVolume(win);
	v.left_volume = (Coefficient)vol[0];
	v.right_volume = (Coefficient)vol[1];
	v.back_left_volume = (Coefficient)vol[2];
	v.back_right_volume = (Coefficient)vol[3];
	v.sample_rate = 0;

	/* Start playing */
	win->yid = YStartPlaySoundObject(con, path, &v);
	if(win->yid != YIDNULL)
	{

	}
	else
	{
	    WinStatusMessage(win, "Play failed", TRUE);
	}

	WinDpyQueueDraw(win);
	WinUpdate(win);
}

/*
 *	Pause.
 */
void WinPause(win_struct *win)
{
	if(win == NULL)
	    return;

	WinStatusMessage(win, "Pausing...", TRUE);

	/* Destroy the playing Sound Object (if any) */
	YDestroyPlaySoundObject(win->con, win->yid);
	win->yid = YIDNULL;                         

	WinStatusMessage(win, "Paused", TRUE);

	WinDpyQueueDraw(win);
	WinUpdate(win);
}

/*
 *	Stop.
 */
void WinStop(win_struct *win)
{
	if(win == NULL)
	    return;

	WinStatusMessage(win, "Stopping...", TRUE);

	/* Destroy the playing Sound Object (if any) */
	YDestroyPlaySoundObject(win->con, win->yid);
	win->yid = YIDNULL;

	/* Rewind */
	WinSetPosition(win, 0l);                         

	WinStatusMessage(win, "Stopped", TRUE);

	WinDpyQueueDraw(win);
	WinUpdate(win);
}

/*
 *	Rewind.
 */
void WinRewind(win_struct *win)
{
	if(win == NULL)
	    return;

	/* Rewind */
	WinSetPosition(win, 0l);

	WinDpyQueueDraw(win);
	WinUpdate(win);
}


/*
 *	Manages the connection to the Y Server and the playing of the
 *	sound object.
 */
void WinManage(win_struct *win)
{
	YConnection *con;
	YID yid;
	YEvent ev;

	if(win == NULL)
	    return;

	con = win->con;
	yid = win->yid;

	/* Not connected to Y Server? */
	if(con == NULL)
	    return;

	/* Manage Y Events */
	while(YGetNextEvent(con, &ev, False) > 0)
	{
	    switch(ev.type)
	    {
	      case YAudioChange:
		break;

	      case YCycleChange:
		break;

	      case YDisconnect:
	      case YShutdown:
		WinConnectionClose(win);
		con = NULL;	/* So we break out of this loop */
		break;

	      case YSoundObjectKill:
		if(ev.kill.yid == yid)
		{
		    WinStop(win);
		    yid = YIDNULL;
		}
		break;

	      case YSHMSound:
		if(ev.yshmsound.minor_op_code == YSHMSoundClose)
		{
		    /* Need to reopen YSHM sound buffer when Audio
		     * mode changes
		     */
		    YSHMCloseSound(con, win->yshm_buf);
		    win->yshm_buf = (guint8 *)YSHMOpenSound(
			con, &win->yshm_buf_len, &win->yshm_id
		    );
		}
		break;

	    }
	}

	/* Still connected and playing? */
	if((con != NULL) && (yid != YIDNULL))
	{
	    YEventSoundPlay v;

	    /* Get current values of the sound object being played */
	    if(!YGetPlaySoundObjectValues(con, yid, &v))
	    {
		const unsigned long flags = v.flags;

		/* Begin updating widgets to reflect current values  */

		win->freeze_count++;

		/* Update position */
		if(flags & YPlayValuesFlagPosition)
		{
		    GtkRange *range = GTK_RANGE(win->position_scale);
		    gtk_adjustment_set_value(
			range->adjustment,
			(gfloat)v.position
		    );
		}

		/* Update volumes */
		if(flags & YPlayValuesFlagVolume)
		{
		    GtkRange *range = GTK_RANGE(win->volume_fleft_scale);
		    gtk_adjustment_set_value(
			range->adjustment,
			1.0f - (gfloat)v.left_volume
		    );
		    range = GTK_RANGE(win->volume_fright_scale);
		    gtk_adjustment_set_value(
			range->adjustment,
			1.0f - (gfloat)v.right_volume
		    );
		    range = GTK_RANGE(win->volume_bleft_scale);
		    gtk_adjustment_set_value(
			range->adjustment,
			1.0f - (gfloat)v.back_left_volume
		    );
		    range = GTK_RANGE(win->volume_bright_scale);
		    gtk_adjustment_set_value(
			range->adjustment,
			1.0f - (gfloat)v.back_right_volume
		    );
		}

		win->freeze_count--;
	    }
	}

	/* Draw the display */
	WinDpyQueueDraw(win);

	/* Draw the DSPView */
	DSPViewDraw(win->dspview, win->yshm_buf, win->yshm_buf_len);
}


/*
 *	Sets the Win's DSPView display style.
 */
void WinDSPViewSetDisplayStyle(
	win_struct *win, dspview_display_style display_style  
)
{
	gint i;
	dspview_display_style cur_style;
	win_dspview_style_struct *sp, styles_list[] = WIN_DSPVIEW_STYLES;
	const gint nstyles = sizeof(styles_list) / sizeof(win_dspview_style_struct);
	dspview_struct *dv = (win != NULL) ? win->dspview : NULL;
	if(dv == NULL)
	    return;

	/* Get current DSPView display style */
	cur_style = dv->display_style;

	/* No change in the DSPView display style? */
	if(display_style == cur_style)
	    return;

	/* Find current DSPView display style in our list of preset
	 * DSPView display styles
	 */
	for(i = 0; i < nstyles; i++)
	{
	    sp = &styles_list[i];
	    if(sp->display_style == display_style)
		break;
	}
	if(i >= nstyles)
	    return;

	sp = &styles_list[i];

	DSPViewSetDisplayStyle(dv, sp->display_style);
	DSPViewSetMergeMode(dv, sp->merge_mode);
	DSPViewSetDSPColor(dv, sp->color);
	DSPViewSetFadeCoeff(dv, sp->fade_coeff);

	WinUpdate(win);
}

/*
 *	Switches the Win's DSPView to the next display style.
 */
void WinDSPViewDisplayStyleNext(win_struct *win)
{
	gint i;
	dspview_display_style cur_style;
	win_dspview_style_struct *sp, styles_list[] = WIN_DSPVIEW_STYLES;
	const gint nstyles = sizeof(styles_list) / sizeof(win_dspview_style_struct);
	dspview_struct *dv = (win != NULL) ? win->dspview : NULL;
	if(dv == NULL)
	    return;

	/* Get current DSPView display style */
	cur_style = dv->display_style;

	/* Find current DSPView display style in our list of preset
	 * DSPView display styles
	 */
	for(i = 0; i < nstyles; i++)
	{
	    sp = &styles_list[i];
	    if(sp->display_style == cur_style)
		break;
	}

	/* Go to "next" style and cycle back to first style as needed */
	i++;
	if(i >= nstyles)
	    i = 0;
	sp = &styles_list[i];

	DSPViewSetDisplayStyle(dv, sp->display_style);
	DSPViewSetMergeMode(dv, sp->merge_mode);
	DSPViewSetDSPColor(dv, sp->color);
	DSPViewSetFadeCoeff(dv, sp->fade_coeff);

	WinUpdate(win);
}

/*
 *	Switches the Win's DSPView to the previous display style.
 */
void WinDSPViewDisplayStylePrev(win_struct *win)
{
	gint i;
	dspview_display_style cur_style;
	win_dspview_style_struct *sp, styles_list[] = WIN_DSPVIEW_STYLES;
	const gint nstyles = sizeof(styles_list) / sizeof(win_dspview_style_struct);
	dspview_struct *dv = (win != NULL) ? win->dspview : NULL;
	if(dv == NULL)
	    return;

	/* Get current DSPView display style */
	cur_style = dv->display_style;

	/* Find current DSPView display style in our list of preset
	 * DSPView display styles
	 */
	for(i = 0; i < nstyles; i++)
	{
	    sp = &styles_list[i];
	    if(sp->display_style == cur_style)
		break;
	}    

	/* Go to "previous" style and cycle to last style as needed */
	i--;
	if(i < 0)
	    i = nstyles - 1;
	sp = &styles_list[i];

	DSPViewSetDisplayStyle(dv, sp->display_style);
	DSPViewSetMergeMode(dv, sp->merge_mode);
	DSPViewSetDSPColor(dv, sp->color);
	DSPViewSetFadeCoeff(dv, sp->fade_coeff);

	WinUpdate(win);
}


/*
 *	Creates a new Win OPID.
 */
win_opid_struct *WinOPIDNew(
	win_opid id, win_struct *win
)
{
	win_opid_struct *opid = WIN_OPID(
	    g_malloc0(sizeof(win_opid_struct))
	);
	opid->id = id;
	opid->win = win;
	return(opid);
}

/*
 *	Deletes the Win OPID.
 */
void WinOPIDDelete(win_opid_struct *opid)
{
	if(opid == NULL)
	    return;

	g_free(opid->button_name);
	g_free(opid->menu_name);
	g_free(opid->tooltip);
	g_free(opid);
}

/*
 *	Creates a new Win OPID list.
 */
GList *WinOPIDListNew(win_struct *win)
{
	gboolean allow_multiple;
	toolbar_item_type item_type;
	const gchar	*button_name,
			*menu_name,
			*tooltip;
	guint	accel_key,
		accel_mods;
	guint8	**button_icon_data,
		**menu_icon_data;
	void (*func_cb)(GtkWidget *, gpointer);
	GList *glist = NULL;

#define DO_APPEND(_id_)	{			\
 win_opid_struct *opid = WinOPIDNew(		\
  (_id_), win					\
 );						\
 opid->allow_multiple = allow_multiple;		\
 opid->item_type = item_type;			\
 opid->button_name = STRDUP(button_name);	\
 opid->menu_name = STRDUP(menu_name);		\
 opid->tooltip = STRDUP(tooltip);		\
 opid->accel_key = accel_key;			\
 opid->accel_mods = accel_mods;			\
 opid->button_icon_data = button_icon_data;	\
 opid->menu_icon_data = menu_icon_data;		\
 opid->func_cb = func_cb;			\
 opid->enter_func_cb = WinOPIDEnterEventCB;	\
 opid->leave_func_cb = WinOPIDLeaveEventCB;	\
 glist = g_list_append(glist, opid);		\
}

	allow_multiple = TRUE;
	item_type = TOOLBAR_ITEM_SEPARATOR;
	button_name = NULL;
	menu_name = button_name;
	tooltip = NULL;
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = NULL;
	DO_APPEND(WIN_OPID_SEPARATOR)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Open";
	menu_name = "Open...";
	tooltip = "Open sound object";
	accel_key = GDK_o;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_open_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinOpenCB;
	DO_APPEND(WIN_OPID_OPEN)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "OpenPlay";
	menu_name = "Play On Open";
	tooltip = "Toggle the play on open";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinPlayOnOpenCB;
	DO_APPEND(WIN_OPID_PLAY_ON_OPEN)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Close";
	menu_name = "Close";
	tooltip = "Close this window";
	accel_key = GDK_w;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_close_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinCloseCB;
	DO_APPEND(WIN_OPID_CLOSE)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Play";
	menu_name = "Play";
	tooltip = "Start playing the sound object";
	accel_key = GDK_Return;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_yplayer_play_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinPlayCB;
	DO_APPEND(WIN_OPID_PLAY)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Pause";
	menu_name = "Pause";
	tooltip = "Pause the playing of the sound object";
	accel_key = GDK_p;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_yplayer_pause_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinPauseCB;
	DO_APPEND(WIN_OPID_PAUSE)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Stop";
	menu_name = "Stop";
	tooltip = "Stop playing the sound object";
	accel_key = GDK_s;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_yplayer_stop_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinStopCB;
	DO_APPEND(WIN_OPID_STOP)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Rewind";
	menu_name = "Rewind";
	tooltip = "Rewind to beginning";
	accel_key = GDK_Home;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_yplayer_rewind_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinRewindCB;
	DO_APPEND(WIN_OPID_REWIND)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "Repeat";
	menu_name = "Repeating";
	tooltip = "Toggle repeating playback";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinRepeatingCB;
	DO_APPEND(WIN_OPID_REPEATING)

	allow_multiple = FALSE;   
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "AutoAudio";
	menu_name = "Auto Change Audio Mode";
	tooltip = "Toggle automatic change of the Audio mode as needed";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinAutoChangeAudioModeCB;
	DO_APPEND(WIN_OPID_AUTO_CHANGE_AUDIO_MODE)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Recorder";
	menu_name = "Recorder...";
	tooltip = "Connect to Y Server";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_yserver_location_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinRecorderCB;
	DO_APPEND(WIN_OPID_RECORDER)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Refresh";
	menu_name = button_name;
	tooltip = "Refresh";
	accel_key = GDK_F5;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_reload_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinRefreshCB;
	DO_APPEND(WIN_OPID_REFRESH)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "ToolBar";
	menu_name = "Tool Bar";
	tooltip = "Show/hide the tool bar";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_tool_bar_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinViewToolBarCB;
	DO_APPEND(WIN_OPID_VIEW_TOOLBAR)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "Display";
	menu_name = "Display";
	tooltip = "Show/hide the display";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = NULL;
	func_cb = WinViewDisplayCB;
	DO_APPEND(WIN_OPID_VIEW_DISPLAY)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "Volume";
	menu_name = "Volume";
	tooltip = "Show/hide the volume";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = NULL;
	func_cb = WinViewVolumeCB;
	DO_APPEND(WIN_OPID_VIEW_VOLUME)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "StatusBar";
	menu_name = "Status Bar";
	tooltip = "Show/hide the status bar";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_status_bar_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinViewStatusBarCB;
	DO_APPEND(WIN_OPID_VIEW_STATUSBAR)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "Wave";
	menu_name = "Wave";
	tooltip = "Set the DSPView style to \"Wave\"";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinDSPViewStyleWaveCB;
	DO_APPEND(WIN_OPID_DSPVIEW_STYLE_WAVE)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "SGraph";
	menu_name = "Sine Graph";
	tooltip = "Set the DSPView style to \"Sine Graph\"";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinDSPViewStyleSineGraphCB;
	DO_APPEND(WIN_OPID_DSPVIEW_STYLE_SINE_GRAPH)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "Analyzer";
	menu_name = "Analyzer";
	tooltip = "Set the DSPView style to \"Analyzer\"";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinDSPViewStyleAnalyzerCB;  
	DO_APPEND(WIN_OPID_DSPVIEW_STYLE_ANALYZER)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "Lghting";
	menu_name = "Lightening";
	tooltip = "Set the DSPView style to \"Lightening\"";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinDSPViewStyleLighteningCB;
	DO_APPEND(WIN_OPID_DSPVIEW_STYLE_LIGHTENING)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "Plasma";
	menu_name = "Plasma";
	tooltip = "Set the DSPView style to \"Plasma\"";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinDSPViewStylePlasmaCB;
	DO_APPEND(WIN_OPID_DSPVIEW_STYLE_PLASMA)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
	button_name = "Nova";
	menu_name = "Nova";
	tooltip = "Set the DSPView style to \"Nova\"";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinDSPViewStyleNovaCB;
	DO_APPEND(WIN_OPID_DSPVIEW_STYLE_NOVA)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Sty Next";
	menu_name = "Next DSPView Style";
	tooltip = "Change to the next DSPView style";
	accel_key = ']';
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinDSPViewStyleNextCB;
	DO_APPEND(WIN_OPID_DSPVIEW_STYLE_NEXT)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Sty Prev";
	menu_name = "Previous DSPView Style";
	tooltip = "Change to the previous DSPView style";
	accel_key = '[';
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;                 
	func_cb = WinDSPViewStylePrevCB;
	DO_APPEND(WIN_OPID_DSPVIEW_STYLE_PREV)


	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Player";
	menu_name = "New Player";
	tooltip = "Create a new Player";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_yplayer_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinWindowsNewPlayerCB;
	DO_APPEND(WIN_OPID_WINDOWS_NEW_PLAYER)

	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "Mixers";
	menu_name = "Mixers";
	tooltip = "Open Mixers";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_ymixer_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinWindowsMixersCB;
	DO_APPEND(WIN_OPID_WINDOWS_MIXERS)


	allow_multiple = FALSE;
	item_type = TOOLBAR_ITEM_BUTTON;
	button_name = "About";          
	menu_name = "About...";
	tooltip = "About this application";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_about_20x20_xpm;
	menu_icon_data = button_icon_data;                 
	func_cb = WinHelpAboutCB;
	DO_APPEND(WIN_OPID_ABOUT)


#undef DO_APPEND

	return(glist);
}

/*
 *	Deletes the Win OPID list.
 */
void WinOPIDListDelete(GList *glist)
{
	if(glist == NULL)
	    return;

	g_list_foreach(glist, (GFunc)WinOPIDDelete, NULL);
	g_list_free(glist);
}

/*
 *	Returns the Win OPID that matches the specified id in the
 *	list.
 */
win_opid_struct *WinOPIDListFind(GList *glist, win_opid id)
{
	win_opid_struct *opid;

	for(;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    opid = WIN_OPID(glist->data);
	    if((opid != NULL) ? (opid->id == id) : FALSE)
		return(opid);
	}

	return(NULL);
}



/*
 *	Builds the Win's Menu Bar.
 */
static void WinBuildMenuBar(win_struct *win)
{
	guint8 **icon_data;
	GtkAccelGroup *accelgrp = win->accelgrp;
	GtkWidget *w, *parent, *menubar, *menu, *menuitem_func_w;
/*	core_struct *core = win->core; */

	parent = win->menubar_handle;


	/* Recreate Menu Bar */
	GTK_WIDGET_DESTROY(win->menubar)
	win->menubar = menubar = w = GUIMenuBarCreate(NULL);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);

#define DO_ADD_MENU_ITEM_LABEL(_id_)	{	\
 win_opid_struct *opid = WinOPIDListFind(	\
  win->opid, (_id_)				\
 );						\
 if(opid != NULL) {				\
  w = GUIMenuItemCreate(                        \
   menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,	\
   (icon_data != NULL) ?			\
    icon_data : opid->menu_icon_data,		\
   opid->menu_name,				\
   opid->accel_key, opid->accel_mods,		\
   (gpointer *)&menuitem_func_w,		\
   win, opid->func_cb				\
  );                                            \
  GUISetMenuItemCrossingCB(                     \
   w,                                           \
   (gpointer)opid->enter_func_cb, opid,		\
   (gpointer)opid->leave_func_cb, opid		\
  );                                            \
 }                                              \
 else						\
  w = NULL;					\
}

#define DO_ADD_MENU_ITEM_CHECK(_id_)	{	\
 win_opid_struct *opid = WinOPIDListFind(	\
  win->opid, (_id_)				\
 );						\
 if(opid != NULL) {				\
  w = GUIMenuItemCreate(			\
   menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp,	\
   (icon_data != NULL) ?			\
    icon_data : opid->menu_icon_data,		\
   opid->menu_name,				\
   opid->accel_key, opid->accel_mods,		\
   (gpointer *)&menuitem_func_w,		\
   win, opid->func_cb				\
  );						\
  GUISetMenuItemCrossingCB(			\
   w,						\
   (gpointer)opid->enter_func_cb, opid,		\
   (gpointer)opid->leave_func_cb, opid		\
  );						\
 }                                              \
 else                                           \
  w = NULL;                                     \
}

#define DO_ADD_MENU_SEP {                       \
 w = GUIMenuItemCreate(                         \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL,     \
  NULL, NULL, 0, 0, NULL,                       \
  NULL, NULL                                    \
 );                                             \
}

	/* Create file menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_OPEN)
	    win->open_mi = w;

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_CHECK(WIN_OPID_PLAY_ON_OPEN)
	    win->play_on_open_micheck = w;

	    DO_ADD_MENU_SEP

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_CLOSE)
	    win->close_mi = w;
	}
	GUIMenuAddToMenuBar(
	    menubar, menu,
#if defined(PROG_LANGUAGE_SPANISH)
"Archivo"
#elif defined(PROG_LANGUAGE_FRENCH)
"Fichier"
#elif defined(PROG_LANGUAGE_GERMAN)
"Akte"
#elif defined(PROG_LANGUAGE_ITALIAN)
"File"
#elif defined(PROG_LANGUAGE_DUTCH)
"Dossier"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Arquivo"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Arkiv"
#else
"File"
#endif
	    , GUI_MENU_BAR_ALIGN_LEFT
	);

	/* Create Player menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_PLAY)
	    win->play_mi = w;

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_PAUSE)
	    win->pause_mi = w;

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_STOP)
	    win->stop_mi = w;

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_REWIND)
	    win->rewind_mi = w;

	    DO_ADD_MENU_SEP

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_CHECK(WIN_OPID_REPEATING)
	    win->repeating_micheck = w;

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_CHECK(WIN_OPID_AUTO_CHANGE_AUDIO_MODE)
	    win->auto_change_audio_mode_micheck = w;

	    DO_ADD_MENU_SEP

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_RECORDER)
	    win->recorder_mi = w;
	}
	GUIMenuAddToMenuBar( 
	    menubar, menu, "Player", GUI_MENU_BAR_ALIGN_LEFT
	);


	/* Create View menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
	    GtkWidget *submenu;

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_CHECK(WIN_OPID_VIEW_DISPLAY)
	    win->view_display_micheck = w;

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_CHECK(WIN_OPID_VIEW_VOLUME)
	    win->view_volume_micheck = w;

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_CHECK(WIN_OPID_VIEW_STATUSBAR)
	    win->view_statusbar_micheck = w;

	    DO_ADD_MENU_SEP

	    /* DSPView Style Submenu */
	    win->dspview_style_menu = submenu = GUIMenuCreate();
	    if(submenu != NULL)
	    {
		GtkWidget *menu = submenu;

		icon_data = NULL;
		DO_ADD_MENU_ITEM_CHECK(WIN_OPID_DSPVIEW_STYLE_WAVE)
		win->dspview_style_wave_micheck = w;

		icon_data = NULL;
		DO_ADD_MENU_ITEM_CHECK(WIN_OPID_DSPVIEW_STYLE_SINE_GRAPH)
		win->dspview_style_sine_graph_micheck = w;

		icon_data = NULL;
		DO_ADD_MENU_ITEM_CHECK(WIN_OPID_DSPVIEW_STYLE_ANALYZER)
		win->dspview_style_analyzer_micheck = w;

		icon_data = NULL;
		DO_ADD_MENU_ITEM_CHECK(WIN_OPID_DSPVIEW_STYLE_LIGHTENING)
		win->dspview_style_lightening_micheck = w;

		icon_data = NULL;
		DO_ADD_MENU_ITEM_CHECK(WIN_OPID_DSPVIEW_STYLE_PLASMA)
		win->dspview_style_plasma_micheck = w;

		icon_data = NULL;
		DO_ADD_MENU_ITEM_CHECK(WIN_OPID_DSPVIEW_STYLE_NOVA)
		win->dspview_style_nova_micheck = w;
	    }
	    win->dspview_style_mi = w = GUIMenuItemCreate(
		menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,
		NULL,
		"DSPView Style",
		0, 0,
		NULL, NULL, NULL
	    );
	    GUIMenuItemSetSubMenu(w, submenu);

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_DSPVIEW_STYLE_NEXT)
	    win->dspview_style_next_mi = w;

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_DSPVIEW_STYLE_PREV)
	    win->dspview_style_prev_mi = w;
	}
	GUIMenuAddToMenuBar(
	    menubar, menu,   
#if defined(PROG_LANGUAGE_SPANISH)
"Vista"
#elif defined(PROG_LANGUAGE_FRENCH)
"Vue"
#elif defined(PROG_LANGUAGE_GERMAN)
"Blick"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Veduta"
#elif defined(PROG_LANGUAGE_DUTCH)
"Overzicht"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Vista"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Sikt"
#else
"View"
#endif
	    , GUI_MENU_BAR_ALIGN_LEFT
	);


	/* Create Windows menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_WINDOWS_NEW_PLAYER)
	    win->windows_new_player_mi = w;

	    DO_ADD_MENU_SEP

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_WINDOWS_MIXERS)
	    win->windows_mixers_mi = w;
	}
	GUIMenuAddToMenuBar(
	    menubar, menu,
#if defined(PROG_LANGUAGE_SPANISH)
"Ventanas"
#elif defined(PROG_LANGUAGE_FRENCH)
"Windows"
#elif defined(PROG_LANGUAGE_GERMAN)
"Fenster"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Finestre"
#elif defined(PROG_LANGUAGE_DUTCH)
"Ramen"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Janelas"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Vinduer"
#else
"Windows"
#endif
	    , GUI_MENU_BAR_ALIGN_LEFT
	);


	/* Create Help menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{

	    icon_data = NULL;
	    DO_ADD_MENU_ITEM_LABEL(WIN_OPID_ABOUT)
	    win->about_mi = w;
	}
	GUIMenuAddToMenuBar( 
	    menubar, menu,
#if defined(PROG_LANGUAGE_SPANISH)
"Ayuda"
#elif defined(PROG_LANGUAGE_FRENCH)
"Aide"
#elif defined(PROG_LANGUAGE_GERMAN)
"Hilfe"
#elif defined(PROG_LANGUAGE_ITALIAN)
"L'Aiuto"
#elif defined(PROG_LANGUAGE_DUTCH)
"Hulp"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Ajuda"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Hjelp"
#else
"Help"
#endif
	    , GUI_MENU_BAR_ALIGN_RIGHT
	);

#undef DO_ADD_MENU_ITEM_LABEL
#undef DO_ADD_MENU_ITEM_CHECK
#undef DO_ADD_MENU_SEP
}

/*
 *	Builds the Win's Tool Bar.
 */
static void WinBuildToolBar(win_struct *win)
{
	GList *glist, *tb_cfg;
	toolbar_item_struct **tb_item = NULL;
	gint tb_total_items = 0;
	toolbar_display tb_display = TOOLBAR_DISPLAY_PICTURES_AND_TEXT;
	toolbar_relief tb_relief = TOOLBAR_RELIEF_NONE;
	GtkWidget *parent;
	toolbar_struct *tb;
	win_opid_struct *opid;
/*	core_struct *core = win->core; */

	parent = win->toolbar_handle;


	/* Create Tool Bar configuration */
	glist = NULL;
	glist = g_list_append(
	    glist, (gpointer)WIN_OPID_CLOSE
	);
	glist = g_list_append(
	    glist, (gpointer)WIN_OPID_SEPARATOR
	);
	glist = g_list_append(
	    glist, (gpointer)WIN_OPID_REFRESH
	);
	glist = g_list_append(
	    glist, (gpointer)WIN_OPID_VIEW_STATUSBAR
	);
	tb_cfg = glist;

	/* Create Tool Bar Items list from the Tool Bar configuration */
	for(glist = tb_cfg;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    /* Get Operation ID from the Tool Bar configuration which
	     * specifies the ID
	     */
	    opid = WinOPIDListFind(win->opid, (win_opid)glist->data);
	    if(opid == NULL)
		continue;

	    ToolBarItemListAppend(
		&tb_item, &tb_total_items,
		ToolBarItemNew(
		    opid->item_type,		/* Item type */
		    opid->button_name,		/* Label text */
		    opid->button_icon_data,	/* Icon data */
		    opid->tooltip,		/* Tooltip text */
		    opid->id,			/* ID */
		    WinToolBarItemCB, opid,
		    WinToolBarItemEnterCB, opid,
		    WinToolBarItemLeaveCB, opid
		)
	    );
	}


	/* Recreate Tool Bar */
	ToolBarDelete(win->toolbar);
	win->toolbar = tb = ToolBarNew(
	    tb_item, tb_total_items, parent,
	    tb_display, tb_relief, FALSE
	);
	ToolBarMap(tb);

	/* Delete Tool Bar items list */
	ToolBarItemListDeleteAll(&tb_item, &tb_total_items);

	/* Delete Tool Bar configuration */
	g_list_free(tb_cfg);

	/* Need to queue resize since Tool Bar size may have changed */
	gtk_widget_queue_resize(win->toplevel);
}

/*
 *	Builds the Win's Lists.
 */
static void WinBuildLists(win_struct *win)
{
	const gint	border_major = 5,
			border_minor = 2;
	const GtkTargetEntry dnd_tar_types[] = {
		{ "text/plain",		0,	DND_INFO_TEXT_PLAIN },
		{ "text/uri-list",	0,	DND_INFO_TEXT_URI_LIST },
		{ "STRING",		0,	DND_INFO_STRING }
	};
	const guint8 midi_12x12_bm[] = {
0x00, 0x06,
0x80, 0x07,
0xe0, 0x05,
0x78, 0x06,
0x98, 0x05,
0x68, 0x04,
0x18, 0x07,
0x88, 0x07,
0x8e, 0x07,
0x0f, 0x03,
0x0f, 0x00,
0x06, 0x00
	};
	const guint8 play_12x12_bm[] = {
0x00, 0x00,
0x10, 0x00,
0x30, 0x00,
0x70, 0x00,
0xf0, 0x00,
0xf0, 0x01,
0xf0, 0x01,
0xf0, 0x00,
0x70, 0x00,
0x30, 0x00,
0x10, 0x00,
0x00, 0x00
	};
	const guint8 pause_12x12_bm[] = {
0x00, 0x00,
0x0c, 0x03,
0x0c, 0x03,
0x0c, 0x03,
0x0c, 0x03,
0x0c, 0x03,
0x0c, 0x03,
0x0c, 0x03,
0x0c, 0x03,
0x0c, 0x03,
0x0c, 0x03,
0x00, 0x00
	};
	const guint8 stop_12x12_bm[] = {
0x00, 0x00,
0xfe, 0x07,
0xfe, 0x07,
0xfe, 0x07,
0xfe, 0x07,
0xfe, 0x07,
0xfe, 0x07,
0xfe, 0x07,
0xfe, 0x07,
0xfe, 0x07,
0xfe, 0x07,
0x00, 0x00
	};
	const guint8 busy_12x12_bm[] = {
0xfc, 0x03,
0x04, 0x02,
0x08, 0x01,
0x28, 0x01,
0xd0, 0x00,
0x60, 0x00,
0x60, 0x00,
0xb0, 0x00,
0x58, 0x01,
0xa8, 0x01,
0x54, 0x03,
0xfc, 0x03
	};
	GdkColor *c;
	GdkColormap *colormap;
	GdkWindow *window;
	GtkRcStyle *rcstyle;
	GtkWidget *w, *parent, *parent2, *parent3;
	dspview_struct *dv;
	core_struct *core = CORE(win->core);
	cfg_item_struct *cfg_list = core->cfg_list;


	/* Get and set up parent */
	parent = win->lists_vbox;
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);    
	gtk_widget_show(w);                                     
	parent = w;

	/* Display parent hbox */
	win->display_parent = w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	parent2 = w;

	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;	

	win->display_da = w = gtk_drawing_area_new();
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS);
	gtk_widget_set_usize(w, WIN_DISPLAY_WIDTH, WIN_DISPLAY_HEIGHT);
	gtk_widget_add_events(
	    w,
	    GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
	    GDK_FOCUS_CHANGE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "configure_event",
	    GTK_SIGNAL_FUNC(WinDpyEventCB), win
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "expose_event",
	    GTK_SIGNAL_FUNC(WinDpyEventCB), win
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "focus_in_event",
	    GTK_SIGNAL_FUNC(WinDpyEventCB), win
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "focus_out_event",
	    GTK_SIGNAL_FUNC(WinDpyEventCB), win
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(WinDpyEventCB), win
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(WinDpyEventCB), win
	);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_realize(w);
	window = w->window;
	GUIDNDSetTar(
	    w,
	    dnd_tar_types,
	    sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
	    GDK_ACTION_COPY | GDK_ACTION_MOVE |
		GDK_ACTION_LINK,		/* Actions */
	    GDK_ACTION_COPY,			/* Default action if same */
	    GDK_ACTION_COPY,			/* Default action */
	    WinDNDDragDataReceivedCB,
	    win
	);
	gtk_widget_show(w);


	/* Create display drawing resources */
	win->gc = GDK_GC_NEW();

	win->colormap = colormap = gdk_window_get_colormap(window);
	if(colormap != NULL)
	    gdk_colormap_ref(colormap);

	c = &win->color_fg;
	GDK_COLOR_SET_COEFF(c, 0.0f, 1.0f, 0.0f);
	GDK_COLORMAP_ALLOC_COLOR(colormap, c);
	c = &win->color_fg_dark;
	GDK_COLOR_SET_COEFF(c, 0.0f, 0.5f, 0.0f);
	GDK_COLORMAP_ALLOC_COLOR(colormap, c);   

	c = &win->color_bg;
	GDK_COLOR_SET_COEFF(c, 0.0f, 0.0f, 0.0f);
	GDK_COLORMAP_ALLOC_COLOR(colormap, c);

	if(TRUE)
	{
	    const gint nticks = WIN_NVOL_TICKS;
	    const gfloat yellow_threshold = 0.7f;
	    gint i, n, gamma_i_offset;
	    gfloat gamma;

	    /* Set volume tick colors */
	    c = win->color_vol_ticks;
	    n = MIN(nticks * yellow_threshold, nticks);
	    gamma_i_offset = 0;
	    for(i = 0; i < n; i++)
	    {
		gamma = (gfloat)(i + gamma_i_offset) /
		    (gfloat)MAX(n + gamma_i_offset, 1);
		GDK_COLOR_SET_COEFF(
		    &c[i],
		    gamma * 1.0f, 1.0f, 0.0f
		)
	    }
	    n = nticks;
	    gamma_i_offset = -i;
	    for(; i < n; i++)
	    {
		gamma = 1.0f - (
		    (gfloat)(i + gamma_i_offset) / 
		    (gfloat)MAX(n + gamma_i_offset, 1)
		);
		GDK_COLOR_SET_COEFF(
		    &c[i],
		    1.0f, gamma * 1.0f, 0.0f
		)
	    }
	    /* Allocate volume tick colors */
	    GDK_COLORMAP_ALLOC_COLORS(colormap, c, nticks)

	    /* Set dark volume tick colors */
	    c = win->color_vol_ticks_dark;
	    n = MIN(nticks * yellow_threshold, nticks);
	    gamma_i_offset = 0;
	    for(i = 0; i < n; i++)
	    {
		gamma = (gfloat)(i + gamma_i_offset) /
		    (gfloat)MAX(n + gamma_i_offset, 1);
		GDK_COLOR_SET_COEFF(
		    &c[i],
		    gamma * 0.5f, 0.5f, 0.0f
		)
	    }
	    n = nticks;
	    gamma_i_offset = -i;
	    for(; i < n; i++)
	    {
		gamma = 1.0f - (
		    (gfloat)(i + gamma_i_offset) /
		    (gfloat)MAX(n + gamma_i_offset, 1)
		);
		GDK_COLOR_SET_COEFF(
		    &c[i],
		    0.5f, gamma * 0.5f, 0.0f
		)
	    }
	    /* Allocate dark volume tick colors */
	    GDK_COLORMAP_ALLOC_COLORS(colormap, c, nticks)
	}

	win->midi_bm = gdk_bitmap_create_from_data(
	    window, midi_12x12_bm, 12, 12
	);
	win->play_bm = gdk_bitmap_create_from_data(
	    window, play_12x12_bm, 12, 12
	);
	win->pause_bm = gdk_bitmap_create_from_data(
	    window, pause_12x12_bm, 12, 12
	);
	win->stop_bm = gdk_bitmap_create_from_data(
	    window, stop_12x12_bm, 12, 12
	);
	win->busy_bm = gdk_bitmap_create_from_data(
	    window, busy_12x12_bm, 12, 12
	);


	/* DSPView */
	win->dspview = dv = DSPViewNew(
	    parent2,
	    WIN_DSPVIEW_WIDTH, WIN_DSPVIEW_HEIGHT
	);
	if(dv != NULL)
	{
	    gint i;
	    dspview_display_style display_style = (dspview_display_style)CFGItemListGetValueI(
		cfg_list, CFG_PARM_WIN_DISPLAY_STYLE
	    );
	    win_dspview_style_struct *sp, styles_list[] = WIN_DSPVIEW_STYLES;
	    const gint nstyles = sizeof(styles_list) / sizeof(win_dspview_style_struct);

	    w = dv->view_da;

	    /* Find DSPView display style in our list of preset
	     * DSPView display styles
	     */
	    for(i = 0; i < nstyles; i++)
	    {
		sp = &styles_list[i];
		if(sp->display_style == display_style)
		{
		    DSPViewSetDisplayStyle(dv, sp->display_style);
		    DSPViewSetMergeMode(dv, sp->merge_mode);
		    DSPViewSetDSPColor(dv, sp->color);
		    DSPViewSetFadeCoeff(dv, sp->fade_coeff);
		    break;
		}
	    }

	    gtk_widget_add_events(
		w,
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(WinDSPViewEventCB), win
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(WinDSPViewEventCB), win
	    );

	    DSPViewMap(dv);
	}

	/* Hbox for Position */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Position Scale */
	win->position_scale = w = gtk_hscale_new(NULL);
        gtk_signal_connect_after(
            GTK_OBJECT(w), "button_press_event",
            GTK_SIGNAL_FUNC(WinScaleEventCB), win
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "button_release_event",
            GTK_SIGNAL_FUNC(WinScaleEventCB), win
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "motion_notify_event",
            GTK_SIGNAL_FUNC(WinScaleEventCB), win
        );
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_scale_set_draw_value(GTK_SCALE(w), FALSE); 
	gtk_signal_connect(
	    GTK_OBJECT(GTK_RANGE(w)->adjustment), "value_changed",
	    GTK_SIGNAL_FUNC(WinPositionValueChangedCB), win
	);
	gtk_widget_show(w);


	/* Hbox for buttons */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

#define SET_CALLBACKS(_w_,_op_)	{		\
 win_opid_struct *opid = WinOPIDListFind(	\
  win->opid, (_op_)				\
 );						\
 if(opid != NULL) {				\
  gtk_signal_connect(				\
   GTK_OBJECT(_w_), "clicked",			\
   GTK_SIGNAL_FUNC(opid->func_cb), opid->win	\
  );						\
  gtk_signal_connect(				\
   GTK_OBJECT(_w_), "enter_notify_event",	\
   GTK_SIGNAL_FUNC(opid->enter_func_cb), opid	\
  );						\
  gtk_signal_connect(				\
   GTK_OBJECT(_w_), "leave_notify_event",	\
   GTK_SIGNAL_FUNC(opid->leave_func_cb), opid	\
  );						\
 }						\
}

	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Open */
	win->open_btn = w = GUIButtonPixmap(
	    (guint8 **)icon_yplayer_eject_14x14_xpm
	); 
	gtk_widget_set_usize(
	    w, WIN_BTN_WIDTH, WIN_BTN_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	SET_CALLBACKS(w, WIN_OPID_OPEN)
	GUISetWidgetTip(w, "Open...");
	gtk_widget_show(w);


	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Play */
	win->play_btn = w = GUIButtonPixmap(
	    (guint8 **)icon_yplayer_play_14x14_xpm
	);
	gtk_widget_set_usize(
	    w, WIN_BTN_WIDTH, WIN_BTN_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	SET_CALLBACKS(w, WIN_OPID_PLAY)
	GUISetWidgetTip(w, "Play");
	gtk_widget_show(w);

	/* Pause */
	win->pause_btn = w = GUIButtonPixmap(
	    (guint8 **)icon_yplayer_pause_14x14_xpm
	);
	gtk_widget_set_usize(
	    w, WIN_BTN_WIDTH, WIN_BTN_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	SET_CALLBACKS(w, WIN_OPID_PAUSE)
	GUISetWidgetTip(w, "Pause");
	gtk_widget_show(w);

	/* Stop */
	win->stop_btn = w = GUIButtonPixmap(
	    (guint8 **)icon_yplayer_stop_14x14_xpm
	);
	gtk_widget_set_usize(
	    w, WIN_BTN_WIDTH, WIN_BTN_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	SET_CALLBACKS(w, WIN_OPID_STOP)
	GUISetWidgetTip(w, "Stop");
	gtk_widget_show(w);

	/* Rewind */
	win->rewind_btn = w = GUIButtonPixmap(
	    (guint8 **)icon_yplayer_rewind_14x14_xpm
	);
	gtk_widget_set_usize(
	    w, WIN_BTN_WIDTH, WIN_BTN_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	SET_CALLBACKS(w, WIN_OPID_REWIND)
	GUISetWidgetTip(w, "Rewind");
	gtk_widget_show(w);



	/* Volume parent vbox */
	parent = win->lists_vbox;
	win->volume_parent = w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);      
	parent2 = w;

	/* Icons hbox */
	w = gtk_hbox_new(TRUE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;
	/* Left speaker icon */
	w = gtk_pixmap_new_from_xpm_d(
	    NULL, NULL, (guint8 **)icon_speaker_left_20x20_xpm
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Right speaker icon */
	w = gtk_pixmap_new_from_xpm_d(
	    NULL, NULL, (guint8 **)icon_speaker_right_20x20_xpm
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Volume scales hbox */
	w = gtk_hbox_new(TRUE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

#define CREATE_VOLUME_SCALE(_parent_,_s_) {	\
 GtkAdjustment *_adj;				\
 GtkWidget *_scale, *_parent2;			\
						\
 /* Create vbox for scale and label */		\
 w = gtk_vbox_new(FALSE, 0);			\
 gtk_box_pack_start(				\
  GTK_BOX(_parent_), w, FALSE, FALSE, 0		\
 );						\
 gtk_widget_show(w);				\
 _parent2 = w;					\
						\
 /* Create adjustment and vscale */		\
 _adj = (GtkAdjustment *)gtk_adjustment_new(	\
  0.0f, 0.0f, 1.0f, 0.1f, 0.1f, 0.0f		\
 );						\
 _scale = w = gtk_vscale_new(_adj);		\
 gtk_signal_connect_after(			\
  GTK_OBJECT(w), "button_press_event",		\
  GTK_SIGNAL_FUNC(WinScaleEventCB), win		\
 );						\
 gtk_signal_connect_after(			\
  GTK_OBJECT(w), "button_release_event",	\
  GTK_SIGNAL_FUNC(WinScaleEventCB), win		\
 );						\
 gtk_signal_connect_after(			\
  GTK_OBJECT(w), "motion_notify_event",		\
  GTK_SIGNAL_FUNC(WinScaleEventCB), win		\
 );						\
 gtk_box_pack_start(				\
  GTK_BOX(_parent2), w, TRUE, TRUE, 0		\
 );						\
 gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);	\
 gtk_signal_connect(				\
  GTK_OBJECT(_adj), "value_changed",		\
  GTK_SIGNAL_FUNC(WinVolumeValueChangedCB), win	\
 );						\
 gtk_widget_show(w);				\
						\
 /* Create label */				\
 if((_s_) != NULL) {				\
  w = gtk_label_new(_s_);			\
  gtk_box_pack_start(				\
   GTK_BOX(_parent2), w, FALSE, FALSE, 0	\
  );						\
  gtk_widget_modify_style(w, rcstyle);		\
  gtk_widget_show(w);				\
 }						\
						\
 w = _scale;					\
}

	rcstyle = gtk_rc_style_new();
	rcstyle->font_name = STRDUP(
"-adobe-helvetica-medium-r-normal-*-8-*-*-*-p-*-iso8859-1"
	);

	CREATE_VOLUME_SCALE(parent3, "B");
	win->volume_bleft_scale = w;

	CREATE_VOLUME_SCALE(parent3, "F");
	win->volume_fleft_scale = w;  

	CREATE_VOLUME_SCALE(parent3, "F");
	win->volume_fright_scale = w;  

	CREATE_VOLUME_SCALE(parent3, "B");
	win->volume_bright_scale = w;

	GTK_RC_STYLE_UNREF(rcstyle);

#undef CREATE_VOLUME_SCALE



#undef SET_CALLBACKS
}

/*
 *	Builds the Win's Status Bar.
 */
static void WinBuildStatusBar(win_struct *win)
{
	gint attach_x = 0;
/*	GtkAdjustment *adj; */
	GtkWidget *w, *parent, *parent2, *parent3;

	parent = win->main_vbox;


	/* Begin recreating status bar */
	GTK_WIDGET_DESTROY(win->statusbar)

	/* Main outer frame */
	win->statusbar = w = gtk_frame_new(NULL);
	gtk_widget_set_usize(w, -1, WIN_STATUS_BAR_HEIGHT);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
	gtk_box_pack_end(GTK_BOX(parent), w, FALSE, FALSE, 0);
	parent = w;

	/* Main table */
	w = gtk_table_new(1, 2, FALSE);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

#if 0
	/* Progress bar */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, 1.0f, 100.0f,
	    0.0f, 0.0f, 0.0f
	);
	win->statusbar_progress = w = gtk_progress_bar_new_with_adjustment(adj);
	gtk_widget_set_usize(w, 100, 20);
	gtk_progress_bar_set_orientation(
	    GTK_PROGRESS_BAR(w), GTK_PROGRESS_LEFT_TO_RIGHT
	);
	gtk_progress_bar_set_bar_style(
	    GTK_PROGRESS_BAR(w), GTK_PROGRESS_CONTINUOUS
	);
	gtk_progress_set_activity_mode(
	    GTK_PROGRESS(w), FALSE
	);
	gtk_table_attach(
	    GTK_TABLE(parent), w,
	    attach_x, attach_x + 1,
	    0, 1,
	    0,
	    GTK_SHRINK,
	    0, 0
	);
	attach_x++;
	gtk_widget_show(w);
#endif

	/* Label */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_container_border_width(GTK_CONTAINER(w), 1);
	gtk_table_attach(
	    GTK_TABLE(parent), w,
	    attach_x, attach_x + 1,
	    0, 1,
	    GTK_SHRINK | GTK_EXPAND | GTK_FILL,
	    GTK_FILL,
	    0, 0
	);
	attach_x++;
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent3 = w;

	win->statusbar_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 2);
	gtk_widget_show(w);
}


/*
 *	Builds the Win's right click menu.
 */
static void WinBuildListsMenu(win_struct *win)
{
	guint8 **icon_data;
	GtkAccelGroup *accelgrp = win->accelgrp;
	GtkWidget *w, *menu, *menuitem_func_w;


	/* Create right click menu */
	GTK_WIDGET_DESTROY(win->lists_menu)
	win->lists_menu = menu = GUIMenuCreate();

#define DO_ADD_MENU_ITEM_LABEL(_id_)    {       \
 win_opid_struct *opid = WinOPIDListFind(       \
  win->opid, (_id_)                             \
 );                                             \
 if(opid != NULL) {                             \
  w = GUIMenuItemCreate(                        \
   menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,    \
   (icon_data != NULL) ?                        \
    icon_data : opid->menu_icon_data,           \
   opid->menu_name,                             \
   opid->accel_key, opid->accel_mods,           \
   (gpointer *)&menuitem_func_w,                \
   win, opid->func_cb                           \
  );                                            \
  GUISetMenuItemCrossingCB(                     \
   w,                                           \
   (gpointer)opid->enter_func_cb, opid,         \
   (gpointer)opid->leave_func_cb, opid          \
  );                                            \
 }                                              \
 else                                           \
  w = NULL;                                     \
}

#define DO_ADD_MENU_ITEM_CHECK(_id_)    {       \
 win_opid_struct *opid = WinOPIDListFind(       \
  win->opid, (_id_)                             \
 );                                             \
 if(opid != NULL) {                             \
  w = GUIMenuItemCreate(                        \
   menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp,    \
   (icon_data != NULL) ?                        \
    icon_data : opid->menu_icon_data,           \
   opid->menu_name,                             \
   opid->accel_key, opid->accel_mods,           \
   (gpointer *)&menuitem_func_w,                \
   win, opid->func_cb                           \
  );                                            \
  GUISetMenuItemCrossingCB(                     \
   w,                                           \
   (gpointer)opid->enter_func_cb, opid,         \
   (gpointer)opid->leave_func_cb, opid          \
  );                                            \
 }                                              \
 else                                           \
  w = NULL;                                     \
}

#define DO_ADD_MENU_SEP {                       \
 w = GUIMenuItemCreate(                         \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL,     \
  NULL, NULL, 0, 0, NULL,                       \
  NULL, NULL                                    \
 );                                             \
}

	icon_data = NULL;
	DO_ADD_MENU_ITEM_LABEL(WIN_OPID_PLAY)
	win->rcm_play_mi = w;

	icon_data = NULL;
	DO_ADD_MENU_ITEM_LABEL(WIN_OPID_PAUSE)
	win->rcm_pause_mi = w;

	icon_data = NULL;
	DO_ADD_MENU_ITEM_LABEL(WIN_OPID_STOP)
	win->rcm_stop_mi = w;

	icon_data = NULL;
	DO_ADD_MENU_ITEM_LABEL(WIN_OPID_REWIND)
	win->rcm_rewind_mi = w;

	DO_ADD_MENU_SEP

	icon_data = NULL;
	DO_ADD_MENU_ITEM_CHECK(WIN_OPID_REPEATING)
	win->rcm_repeating_micheck = w;

	icon_data = NULL;
	DO_ADD_MENU_ITEM_CHECK(WIN_OPID_AUTO_CHANGE_AUDIO_MODE)
	win->rcm_auto_change_audio_mode_micheck = w; 

	DO_ADD_MENU_SEP

	icon_data = (guint8 **)icon_yplayer_eject_20x20_xpm;
	DO_ADD_MENU_ITEM_LABEL(WIN_OPID_OPEN)
	win->rcm_open_mi = w;

	icon_data = NULL;
	DO_ADD_MENU_ITEM_CHECK(WIN_OPID_PLAY_ON_OPEN)
	win->rcm_play_on_open_micheck = w;

#undef DO_ADD_MENU_ITEM_LABEL
#undef DO_ADD_MENU_ITEM_CHECK
#undef DO_ADD_MENU_SEP
}



/*
 *	Creates a new Win.
 */
win_struct *WinNew(gpointer core_ptr)
{
	const gint border_major = 5;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2;
	GtkWidget *toplevel, *main_vbox;
	core_struct *core = CORE(core_ptr);
	const cfg_item_struct *cfg_list = core->cfg_list;
	win_struct *win = WIN(
	    g_malloc0(sizeof(win_struct))
	);
	if((win == NULL) || (core == NULL))
	{
	    g_free(win);
	    return(NULL);
	}

	win->toplevel = toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);  
	win->accelgrp = accelgrp = gtk_accel_group_new();
	win->processing = FALSE;
	win->busy_count = 0;
	win->freeze_count = 0;
	win->core = core_ptr;
	win->opid = WinOPIDListNew(win);
	win->menubar_map_state = TRUE;
	win->toolbar_map_state = TRUE;
	win->display_map_state = CFGItemListGetValueI(
	    cfg_list, CFG_PARM_WIN_SHOW_DISPLAY
	);
	win->volume_map_state = CFGItemListGetValueI(
            cfg_list, CFG_PARM_WIN_SHOW_VOLUME
        );
	win->display_pm = NULL;
	win->gc = NULL;
	win->font = gdk_font_load(
"-adobe-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1"
	);
	win->font_bold = gdk_font_load(
"-adobe-helvetica-bold-r-normal-*-12-*-*-*-p-*-iso8859-1"
	);
	win->font_small = gdk_font_load(
"-adobe-helvetica-medium-r-normal-*-8-*-*-*-p-*-iso8859-1"
	);
	win->colormap = NULL;
	win->statusbar_map_state = CFGItemListGetValueI(
            cfg_list, CFG_PARM_WIN_SHOW_STATUS_BAR
        );
	win->statusbar_progress_last = 0.0f;
	win->manage_toid = 0;
	win->recorder = STRDUP(CFGItemListGetValueS(
            cfg_list, CFG_PARM_RECORDER
        ));
	win->con = NULL;
	win->yshm_buf = NULL;
	win->yshm_buf_len = 0;
	win->yshm_id = -1;
	win->cur_path = NULL;
	win->yid = YIDNULL;
	win->sample_rate = 0;
	win->sample_size = 0;
	win->channels = 0;
	win->play_on_open = CFGItemListGetValueI( 
            cfg_list, CFG_PARM_PLAY_ON_OPEN
        );
;	win->repeating =  CFGItemListGetValueI(
            cfg_list, CFG_PARM_REPEATING
        ); 
	win->auto_change_audio_mode = CFGItemListGetValueI(
            cfg_list, CFG_PARM_AUTO_CHANGE_AUDIO
        );
	win->has_changes = FALSE;

	w = toplevel;
	gtk_window_set_wmclass(
	    GTK_WINDOW(w),
	    (core->wm_name != NULL) ? core->wm_name : "window",
	    (core->wm_class != NULL) ? core->wm_class : PROG_NAME
	);
	gtk_window_set_policy(
	    GTK_WINDOW(w),
	    TRUE, TRUE, TRUE
	);
	if(core->geometry != NULL)
	{
	    const GdkGeometryFlags flags = core->geometry_flags;
	    const GdkRectangle *g = core->geometry;
	    if((flags & GDK_GEOMETRY_X) || (flags & GDK_GEOMETRY_Y))
		gtk_widget_set_uposition(w, g->x, g->y);
	}
	gtk_widget_set_usize(w, WIN_WIDTH, WIN_HEIGHT);
	gtk_widget_add_events(
	    w,
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(WinEventCB), win
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(WinEventCB), win
	);
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    GdkGeometry geo;

	    geo.min_width = 100;
	    geo.min_height = 70;
	    geo.base_width = 0;
	    geo.base_height = 0;
	    geo.width_inc = 1;
	    geo.height_inc = 1;
	    gdk_window_set_geometry_hints(
		window, &geo,
		GDK_HINT_MIN_SIZE |  
		GDK_HINT_BASE_SIZE |
		GDK_HINT_RESIZE_INC
	    );

	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_yplayer_48x48_xpm);
	    gtk_window_set_title(
		GTK_WINDOW(w),
		(core->title != NULL) ? core->title : PROG_NAME
	    );
	}
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(WinDeleteEventCB), win
	);
	gtk_container_border_width(GTK_CONTAINER(w), 0);
	gtk_window_add_accel_group(GTK_WINDOW(w), accelgrp);
	parent = w;


	/* Main VBox */
	win->main_vbox = main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), 0);                        
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Build Menu Bar */
	win->menubar_handle = w = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	win->menubar = NULL;
	WinBuildMenuBar(win);

#if 0
	/* Build Tool Bar */
	win->toolbar_handle = w = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	win->toolbar = NULL;
	WinBuildToolBar(win);
#endif

	/* Vbox for the lists */
	win->lists_vbox = w = gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Build lists */
	WinBuildLists(win);


	/* Build status bar */
	win->statusbar = NULL;
	WinBuildStatusBar(win);

	/* Build right click menu */
	win->lists_menu = NULL;
	WinBuildListsMenu(win);

	/* Show/hide menu bar */
	w = win->menubar_handle;
	if(win->menubar_map_state)
	    gtk_widget_show(w);
	else
	    gtk_widget_hide(w);

#if 0
	/* Show/hide tool bar */
	w = win->toolbar_handle;
	if(win->toolbar_map_state)
	    gtk_widget_show(w);
	else
	    gtk_widget_hide(w);
#endif

	/* Show/hide display */
	w = win->display_parent;
	if(win->display_map_state)
	    gtk_widget_show(w);
	else
	    gtk_widget_hide(w);

	/* Show/hide volume */
	w = win->volume_parent;
	if(win->volume_map_state)
	    gtk_widget_show(w);
	else    
	    gtk_widget_hide(w);

	/* Show/hide status bar */
	w = win->statusbar;
	if(win->statusbar_map_state)
	    gtk_widget_show(w);
	else
	    gtk_widget_hide(w);


	/* Set manage timeout callback */
	win->manage_toid = gtk_timeout_add(
	    100,		/* In milliseconds */
	    WinManageTOCB, win
	);


	WinUpdate(win);


	return(win);
}

/*
 *	Updates the Win's widgets to reflect current values.
 */
void WinUpdate(win_struct *win)
{
	gboolean sensitive, toggled, connected, loaded, playing;
	GtkWidget *w;
	dspview_display_style dspview_style;
	dspview_struct *dv;
	toolbar_struct *toolbar;


	if(win == NULL)
	    return;

	win->freeze_count++;

	connected = (win->con != NULL) ? TRUE : FALSE;
	loaded = !STRISEMPTY(win->cur_path) ? TRUE : FALSE;
	playing = (win->yid != YIDNULL) ? TRUE : FALSE;

	toolbar = win->toolbar;
	dv = win->dspview;
	dspview_style = (dv != NULL) ?
	    dv->display_style : DSPVIEW_DISPLAY_STYLE_WAVE;

	/* Position */
	sensitive = (connected && loaded) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->position_scale, sensitive)

	/* Open */
	sensitive = connected;
	GTK_WIDGET_SET_SENSITIVE(win->open_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->rcm_open_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->open_btn, sensitive)
	ToolBarItemSetSensitiveID(
	    toolbar, WIN_OPID_OPEN, sensitive
	);

	/* Play On Open */
	sensitive = TRUE;
	toggled = win->play_on_open;
	w = win->play_on_open_micheck;
	GTK_WIDGET_SET_SENSITIVE(w, sensitive)
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(w, toggled)
	w = win->rcm_play_on_open_micheck;
	GTK_WIDGET_SET_SENSITIVE(w, sensitive)
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(w, toggled)
	ToolBarItemSetSensitiveID(
	    toolbar, WIN_OPID_PLAY_ON_OPEN, sensitive
	);
	ToolBarItemSetToggleID(
	    toolbar, WIN_OPID_PLAY_ON_OPEN, toggled
	);

	/* Play */
	sensitive = (connected && loaded && !playing) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->play_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->rcm_play_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->play_btn, sensitive)
	ToolBarItemSetSensitiveID(
	    toolbar, WIN_OPID_PLAY, sensitive
	);

	/* Pause */
	sensitive = (connected && loaded && playing) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->pause_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->rcm_pause_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->pause_btn, sensitive)
	ToolBarItemSetSensitiveID(
	    toolbar, WIN_OPID_PAUSE, sensitive
	);

	/* Stop */
	sensitive = (connected && loaded && playing) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->stop_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->rcm_stop_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->stop_btn, sensitive)
	ToolBarItemSetSensitiveID(
	    toolbar, WIN_OPID_STOP, sensitive
	);

	/* Rewind */
	sensitive = (connected && loaded) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->rewind_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->rcm_rewind_mi, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->rewind_btn, sensitive)
	ToolBarItemSetSensitiveID(
	    toolbar, WIN_OPID_REWIND, sensitive
	);

	/* Repeating */
	sensitive = !playing ? TRUE : FALSE;
	toggled = win->repeating;
	w = win->repeating_micheck;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(w, toggled) 
	GTK_WIDGET_SET_SENSITIVE(w, sensitive)
	w = win->rcm_repeating_micheck;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(w, toggled)
	GTK_WIDGET_SET_SENSITIVE(w, sensitive)
	ToolBarItemSetSensitiveID(
	    toolbar, WIN_OPID_REPEATING, sensitive
	);
	ToolBarItemSetToggleID(
	    toolbar, WIN_OPID_REPEATING, toggled
	);

	/* Auto Change Audio Mode */
	sensitive = !playing ? TRUE : FALSE;
	toggled = win->auto_change_audio_mode;
	w = win->auto_change_audio_mode_micheck;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(w, toggled)
	GTK_WIDGET_SET_SENSITIVE(w, sensitive)
	w = win->rcm_auto_change_audio_mode_micheck;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(w, toggled)
	GTK_WIDGET_SET_SENSITIVE(w, sensitive)
	ToolBarItemSetSensitiveID(
	    toolbar, WIN_OPID_AUTO_CHANGE_AUDIO_MODE, sensitive
	);
	ToolBarItemSetToggleID(
	    toolbar, WIN_OPID_AUTO_CHANGE_AUDIO_MODE, toggled
	);

	/* Volume */
	sensitive = (connected && loaded) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->volume_fleft_scale, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->volume_fright_scale, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->volume_bleft_scale, sensitive)
	GTK_WIDGET_SET_SENSITIVE(win->volume_bright_scale, sensitive)


	/* Connect to Y Server */
	sensitive = !playing ? TRUE : FALSE;
	w = win->recorder_mi;
	GTK_WIDGET_SET_SENSITIVE(w, sensitive)


	/* View Tool Bar */
	toggled = win->toolbar_map_state;
	ToolBarItemSetToggleID(
	    toolbar, WIN_OPID_VIEW_TOOLBAR, toggled
	);

	/* View Display */
	toggled = win->display_map_state;
	w = win->view_display_micheck;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(w, toggled) 
	ToolBarItemSetToggleID(
	    toolbar, WIN_OPID_VIEW_DISPLAY, toggled
	);

	/* View Volume */
	toggled = win->volume_map_state;
	w = win->view_volume_micheck;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(w, toggled)
	ToolBarItemSetToggleID(
	    toolbar, WIN_OPID_VIEW_VOLUME, toggled
	);

	/* View Status Bar */
	toggled = win->statusbar_map_state;
	w = win->view_statusbar_micheck;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(w, toggled) 
	ToolBarItemSetToggleID(
	    toolbar, WIN_OPID_VIEW_STATUSBAR, toggled
	);

	/* DSPView Styles */
	toggled = (dspview_style == DSPVIEW_DISPLAY_STYLE_WAVE) ? TRUE : FALSE;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(
	    win->dspview_style_wave_micheck, toggled
	);
	toggled = (dspview_style == DSPVIEW_DISPLAY_STYLE_SGRAPH) ? TRUE : FALSE;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(
	    win->dspview_style_sine_graph_micheck, toggled
	);
	toggled = (dspview_style == DSPVIEW_DISPLAY_STYLE_ANALYZER) ? TRUE : FALSE;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(
	    win->dspview_style_analyzer_micheck, toggled
	);
	toggled = (dspview_style == DSPVIEW_DISPLAY_STYLE_LIGHTENING) ? TRUE : FALSE;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(
	    win->dspview_style_lightening_micheck, toggled
	);
	toggled = (dspview_style == DSPVIEW_DISPLAY_STYLE_PLASMA) ? TRUE : FALSE;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(
	    win->dspview_style_plasma_micheck, toggled
	);
	toggled = (dspview_style == DSPVIEW_DISPLAY_STYLE_NOVA) ? TRUE : FALSE;
	GTK_CHECK_MENU_ITEM_SET_ACTIVE(
	    win->dspview_style_nova_micheck, toggled
	);


	win->freeze_count--;
}

/*
 *	Sets the Win as busy or ready.
 */
void WinSetBusy(win_struct *win, gboolean busy)
{
	GdkCursor *cur;
	GdkWindow *window;
	core_struct *core;
	GtkWidget *w = (win != NULL) ? win->toplevel : NULL;
	if(w == NULL)
	    return;

	core = win->core;

	if(busy)
	{
	    /* Increase busy count */
	    win->busy_count++;

	    /* Already busy? */
	    if(win->busy_count > 1)
		return;

	    cur = core->busy_cur;
	}
	else
	{
	    /* Reduce busy count */
	    win->busy_count--;
	    if(win->busy_count < 0)
		win->busy_count = 0;

	    /* Still busy? */
	    if(win->busy_count > 0)
		return;

	    cur = NULL;
	}
	 
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_cursor(window, cur);
	    gdk_flush();
	}
}


/*                          
 *	Checks if the Win's Tool Bar is shown.      
 */
gboolean WinToolBarIsShown(win_struct *win)
{
	return((win != NULL) ? win->toolbar_map_state : FALSE);
}

/*
 *	Show/hide the Win's Tool Bar.
 */
void WinToolBarSetShow(win_struct *win, gboolean show)
{
	GtkWidget *w = (win != NULL) ? win->toolbar_handle : NULL;
	if(w == NULL)
	    return;

	win->freeze_count++;

	if(win->toolbar_map_state && !show)
	{
	    gtk_widget_hide(w);
	    win->toolbar_map_state = FALSE;
	}
	else if(!win->toolbar_map_state && show)
	{
	    gtk_widget_show(w);
	    win->toolbar_map_state = TRUE;
	}

	WinUpdate(win);

	win->freeze_count--;
}

/*
 *	Recreates the Win's Tool Bar Items.
 */
void WinToolBarUpdateItems(win_struct *win)
{
	if(win == NULL)
	    return;

	/* Recreate Tool Bar and the Tool Bar Items */
	WinBuildToolBar(win);
}

/*
 *	Checks if the Win's Display is shown.
 */
gboolean WinDisplayIsShown(win_struct *win)
{
	return((win != NULL) ? win->display_map_state : FALSE);
}

/*
 *	Show/hide the Win's Display.
 */
void WinDisplaySetShow(win_struct *win, gboolean show)
{
	GtkWidget *w = (win != NULL) ? win->display_parent : NULL;
	if(w == NULL)
	    return;

	win->freeze_count++;

	if(win->display_map_state && !show)
	{
	    gtk_widget_hide(w);
	    win->display_map_state = FALSE;
	}
	else if(!win->display_map_state && show)
	{
	    gtk_widget_show(w);
	    win->display_map_state = TRUE;
	}                                   
	 
	WinUpdate(win);

	win->freeze_count--;
}

/*
 *	Checks if the Win's Volume is shown.
 */
gboolean WinVolumeIsShown(win_struct *win)
{
	return((win != NULL) ? win->volume_map_state : FALSE);
}

/*
 *      Show/hide the Win's Volume.
 */
void WinVolumeSetShow(win_struct *win, gboolean show)
{             
	GtkWidget *w = (win != NULL) ? win->volume_parent : NULL;
	if(w == NULL)                                        
	    return;

	win->freeze_count++;

	if(win->volume_map_state && !show)
	{
	    gtk_widget_hide(w);
	    win->volume_map_state = FALSE;
	}
	else if(!win->volume_map_state && show)
	{
	    gtk_widget_show(w);
	    win->volume_map_state = TRUE;
	}

	WinUpdate(win);

	win->freeze_count--;
}

/*
 *	Checks if the Win's Status Bar is shown.
 */
gboolean WinStatusBarIsShown(win_struct *win)
{
	return((win != NULL) ? win->statusbar_map_state : FALSE);
}

/*
 *	Show/hide the Win's Status Bar.
 */
void WinStatusBarSetShow(win_struct *win, gboolean show)
{
	GtkWidget *w = (win != NULL) ? win->statusbar : NULL;
	if(w == NULL)
	    return;

	win->freeze_count++;

	if(win->statusbar_map_state && !show)
	{
	    gtk_widget_hide(w);
	    win->statusbar_map_state = FALSE;
	}
	else if(!win->statusbar_map_state && show)
	{
	    gtk_widget_show(w);
	    win->statusbar_map_state = TRUE;
	}

	WinUpdate(win);

	win->freeze_count--;
}


/*
 *	Sets the Win's status bar progress.
 */
void WinStatusProgress(
	win_struct *win, gfloat val, gboolean allow_gtk_iteration
)
{
	GtkProgress *progress;
	GtkWidget *w = (win != NULL) ? win->statusbar_progress : NULL;
	if(w == NULL)
	    return;

	progress = GTK_PROGRESS(w);

	/* Do activity? */
	if(val < 0.0f)
	{
	    GtkAdjustment *adj = progress->adjustment;

	    /* Get new value based on unknown position */
	    val = gtk_progress_get_value(progress) + 1.0f;
	    if(val > adj->upper)
		val = adj->lower;

	    /* Update progress bar for `activity mode' (unknown limit)
	     * and set new value            
	     */
	    gtk_progress_set_activity_mode(progress, TRUE);
	    gtk_progress_set_show_text(progress, FALSE);
	    gtk_progress_set_value(progress, val);

	    /* Reset last progress position to -1.0, indicating that
	     * there is no known last position since we are doing
	     * activity mode
	     */
	    win->statusbar_progress_last = -1.0f;
	}
	else
	{   
	    /* Clip value to 0.0, 1.0 */
	    if(val > 1.0f)
		val = 1.0f;

	    /* Reset last progress position if it is greater than the
	     * given value, implying the progress has warped back to
	     * the beginning
	     */
	    if(win->statusbar_progress_last > val)
		win->statusbar_progress_last = -1.0f;

	    /* Has value not sufficiently changed? (less than 0.01
	     * change in value)
	     */
	    if((win->statusbar_progress_last > 0.0f) &&
	       ((val - win->statusbar_progress_last) < 0.01f)
	    )
		return;

	    /* Update progress */ 
	    gtk_progress_set_activity_mode(progress, FALSE);
	    gtk_progress_set_format_string(
		progress, "%p%%"
	    );
	    /* Display text only if value is positive */
	    gtk_progress_set_show_text(
		progress, (val > 0.0f) ? TRUE : FALSE 
	    );
	    gtk_progress_bar_update(GTK_PROGRESS_BAR(w), val); 

	    /* Record last position */
	    win->statusbar_progress_last = val;
	}

	/* Flush events? */
	while((gtk_events_pending() > 0) && allow_gtk_iteration)
	    gtk_main_iteration();
}

/*
 *	Sets the Win's status bar message.
 */
void WinStatusMessage(
	win_struct *win, const gchar *mesg, gboolean allow_gtk_iteration
)
{
	GtkWidget *w = (win != NULL) ? win->statusbar_label : NULL;
	if(w == NULL)
	    return;

	gtk_label_set_text(
	    GTK_LABEL(w), (mesg != NULL) ? mesg : ""
	);  

	while((gtk_events_pending() > 0) && allow_gtk_iteration)
	    gtk_main_iteration();
}

/*
 *	Checks if the Win is mapped.
 */
gboolean WinIsMapped(win_struct *win)
{
	GtkWidget *w = (win != NULL) ? win->toplevel : NULL;
	return((w != NULL) ? GTK_WIDGET_MAPPED(w) : FALSE);
}

/*
 *	Maps the Win.
 */
void WinMap(win_struct *win)
{
	GtkWidget *w = (win != NULL) ? win->toplevel : NULL;
	if(w == NULL)
	    return;

	gtk_widget_show_raise(w);
}

/*
 *	Unmaps the Win.
 */
void WinUnmap(win_struct *win)
{
	GtkWidget *w = (win != NULL) ? win->toplevel : NULL;
	if(w == NULL)
	    return;

	gtk_widget_hide(w);
}

/*
 *	Deletes the Win.
 */
void WinDelete(win_struct *win)
{
	if(win == NULL)
	    return;

	/* Stop playing sound object and close connection to Y server
	 * as needed
	 */
	WinConnectionClose(win);

	GTK_TIMEOUT_REMOVE(win->manage_toid)
	win->manage_toid = 0;

	/* Lists menu */
	GTK_WIDGET_DESTROY(win->lists_menu)

	/* Menu Bar */
	GTK_WIDGET_DESTROY(win->menubar)
	GTK_WIDGET_DESTROY(win->menubar_handle)

	/* Tool Bar */
	ToolBarDelete(win->toolbar);
	GTK_WIDGET_DESTROY(win->toolbar_handle)

	/* Lists VBox */
	DSPViewDelete(win->dspview);
	GTK_WIDGET_DESTROY(win->display_da)
	GTK_WIDGET_DESTROY(win->display_parent)
	GTK_WIDGET_DESTROY(win->volume_bleft_scale)
	GTK_WIDGET_DESTROY(win->volume_fleft_scale)
	GTK_WIDGET_DESTROY(win->volume_fright_scale)
	GTK_WIDGET_DESTROY(win->volume_bright_scale)
	GTK_WIDGET_DESTROY(win->volume_parent)
	GTK_WIDGET_DESTROY(win->open_btn)
	GTK_WIDGET_DESTROY(win->play_btn)
	GTK_WIDGET_DESTROY(win->pause_btn)
	GTK_WIDGET_DESTROY(win->stop_btn)
	GTK_WIDGET_DESTROY(win->rewind_btn)
	GTK_WIDGET_DESTROY(win->position_scale)

	GTK_WIDGET_DESTROY(win->lists_vbox)

	/* Status Bar */
	GTK_WIDGET_DESTROY(win->statusbar)

	/* Main VBox */
	GTK_WIDGET_DESTROY(win->main_vbox)

	/* Toplevel */
	GTK_WIDGET_DESTROY(win->toplevel)

	/* Accel group */
	GTK_ACCEL_GROUP_UNREF(win->accelgrp)

	/* Display resources */
	GDK_GC_UNREF(win->gc)
	GDK_FONT_UNREF(win->font)
	GDK_FONT_UNREF(win->font_bold)
	GDK_FONT_UNREF(win->font_small)
	GDK_PIXMAP_UNREF(win->display_pm)
	GDK_COLORMAP_FREE_COLOR(win->colormap, &win->color_fg)
	GDK_COLORMAP_FREE_COLOR(win->colormap, &win->color_fg_dark)
	GDK_COLORMAP_FREE_COLOR(win->colormap, &win->color_bg)
	GDK_COLORMAP_FREE_COLORS(
	    win->colormap, win->color_vol_ticks, WIN_NVOL_TICKS
	)
	GDK_COLORMAP_FREE_COLORS(
	    win->colormap, win->color_vol_ticks_dark, WIN_NVOL_TICKS 
	)
	GDK_COLORMAP_UNREF(win->colormap)
	GDK_BITMAP_UNREF(win->midi_bm)
	GDK_BITMAP_UNREF(win->play_bm) 
	GDK_BITMAP_UNREF(win->pause_bm) 
	GDK_BITMAP_UNREF(win->stop_bm) 
	GDK_BITMAP_UNREF(win->busy_bm) 

	/* OPIDs list */
	WinOPIDListDelete(win->opid);

	g_free(win->recorder);
	g_free(win->cur_path);

	g_free(win);
}
