/*
			    Win
 */
#ifndef WIN_H
#define WIN_H

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <gtk/gtk.h>
#include "toolbar.h"
#include "dspview.h"


typedef struct _win_struct              win_struct;
typedef struct _win_opid_struct         win_opid_struct;


#define WIN_NVOL_TICKS		15


/*
 *	Operation IDs:
 */
typedef enum {

	WIN_OPID_SEPARATOR				= 10,

	WIN_OPID_OPEN					= 50,
	WIN_OPID_PLAY_ON_OPEN				= 70,
	WIN_OPID_CLOSE					= 100,

	WIN_OPID_PLAY					= 200,
	WIN_OPID_PAUSE					= 201,
	WIN_OPID_STOP					= 202,
	WIN_OPID_REWIND					= 203,
	WIN_OPID_REPEATING				= 204,
	WIN_OPID_AUTO_CHANGE_AUDIO_MODE			= 205,
	WIN_OPID_RECORDER				= 250,

	WIN_OPID_REFRESH				= 300,
	WIN_OPID_VIEW_TOOLBAR				= 310,
	WIN_OPID_VIEW_DISPLAY				= 311,
	WIN_OPID_VIEW_VOLUME				= 312,
	WIN_OPID_VIEW_STATUSBAR				= 313,
	WIN_OPID_DSPVIEW_STYLE_WAVE			= 360,
	WIN_OPID_DSPVIEW_STYLE_SINE_GRAPH		= 361,
	WIN_OPID_DSPVIEW_STYLE_ANALYZER			= 362,
	WIN_OPID_DSPVIEW_STYLE_LIGHTENING		= 363,
	WIN_OPID_DSPVIEW_STYLE_PLASMA			= 364,
	WIN_OPID_DSPVIEW_STYLE_NOVA			= 365,
	WIN_OPID_DSPVIEW_STYLE_NEXT			= 380,
	WIN_OPID_DSPVIEW_STYLE_PREV			= 381,

	WIN_OPID_WINDOWS_NEW_PLAYER			= 500,
	WIN_OPID_WINDOWS_MIXERS				= 510,

	WIN_OPID_ABOUT					= 1000

} win_opid;


/*
 *	Win:
 */
struct _win_struct {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gboolean	processing;
	gint		busy_count,
			freeze_count;
	gpointer	core;

	GList		*opid;

	GtkWidget	*main_vbox;

	gboolean	menubar_map_state;
	GtkWidget	*menubar_handle,
			*menubar;

	GtkWidget	*open_mi,
			*play_on_open_micheck,
			*close_mi,

			*play_mi,
			*pause_mi,
			*stop_mi,
			*rewind_mi,
			*repeating_micheck,
			*auto_change_audio_mode_micheck,
			*recorder_mi,

			*view_display_micheck,
			*view_volume_micheck,
			*view_statusbar_micheck,
			*dspview_style_mi,
			*dspview_style_menu,
			*dspview_style_wave_micheck,
			*dspview_style_sine_graph_micheck,
			*dspview_style_analyzer_micheck,
			*dspview_style_lightening_micheck,
			*dspview_style_plasma_micheck,
			*dspview_style_nova_micheck,
			*dspview_style_next_mi,
			*dspview_style_prev_mi,

			*windows_new_player_mi,
			*windows_mixers_mi,

			*about_mi;

	gboolean	toolbar_map_state;
	GtkWidget	*toolbar_handle;
	toolbar_struct	*toolbar;

	gboolean	display_map_state,
			volume_map_state;
	GdkPixmap	*display_pm;
	GdkGC		*gc;
	GdkFont		*font,
			*font_bold,
			*font_small;
	GdkColormap	*colormap;
	GdkColor	color_fg,
			color_fg_dark,
			color_bg,
			color_vol_ticks[WIN_NVOL_TICKS],
			color_vol_ticks_dark[WIN_NVOL_TICKS];
	GdkBitmap	*midi_bm,
			*play_bm,
			*pause_bm,
			*stop_bm,
			*busy_bm;
	GtkWidget	*lists_vbox,

			*display_parent,
			*display_da,

			*volume_parent,
			*volume_bleft_scale,
			*volume_fleft_scale,
			*volume_fright_scale,
			*volume_bright_scale,

			*position_scale,

			*open_btn,
			*play_btn,
			*pause_btn,
			*stop_btn,
			*rewind_btn;

	dspview_struct	*dspview;

	gboolean	statusbar_map_state;
	GtkWidget	*statusbar;
	GtkWidget	*statusbar_progress,
			*statusbar_label;
	gfloat		statusbar_progress_last;

	GtkWidget	*lists_menu,		/* Right click menu */
			*rcm_play_mi,
			*rcm_pause_mi,
			*rcm_stop_mi,
			*rcm_rewind_mi,
			*rcm_repeating_micheck,
			*rcm_auto_change_audio_mode_micheck,
			*rcm_open_mi,
			*rcm_play_on_open_micheck;

	/* Manage timeout ID */
	guint		manage_toid;

	/* Y Server information */
	gchar		*recorder;		/* Address to Y Server */
	YConnection	*con;			/* Connection to Y Server */
	guint8		*yshm_buf;		/* Y Shared Sound Buffer */
	gint		yshm_buf_len,		/* Y Shared Sound Buffer Length */
			yshm_id;		/* Y Shared Sound Buffer ID */

	/* Sound object information */
	gchar		*cur_path;		/* Path to Sound Object */
	YID		yid;			/* When not YIDNULL means playing */

	gint		sample_rate,		/* In Hz */
			sample_size,		/* In bits */
			channels;		/* 1, 2, or 4 */

	gboolean	play_on_open,
			repeating,
			auto_change_audio_mode;

	gboolean	has_changes;
};
#define WIN(p)		((win_struct *)(p))


/*
 *      Win Operation ID:
 */
struct _win_opid_struct {

	win_opid	id;
	win_struct	*win;        

	gboolean	allow_multiple;
	toolbar_item_type	item_type;

	gchar		*button_name;
	gchar		*menu_name;
	gchar		*tooltip;

	guint		accel_key,
			accel_mods;

	guint8		**button_icon_data,	/* Shared */
			**menu_icon_data;	/* Shared */

	void (*func_cb)(
		GtkWidget *,		/* GtkWidget */
		gpointer		/* Win */
	);
	gint (*enter_func_cb)(
		GtkWidget *,		/* GtkWidget */
		GdkEventCrossing *,	/* GdkEventCrossing */
		gpointer		/* OPID */
	);
	gint (*leave_func_cb)(
		GtkWidget *,		/* GtkWidget */
		GdkEventCrossing *,	/* GdkEventCrossing */
		gpointer		/* OPID */             
	);

};
#define WIN_OPID(p)		((win_opid_struct *)(p))


extern void WinConnectionOpen(win_struct *win, const gchar *address);
extern void WinConnectionClose(win_struct *win);

extern gboolean WinIsPlaying(win_struct *win);
extern gulong WinGetPosition(win_struct *win);
extern void WinSetPosition(win_struct *win, gulong pos);
extern gfloat *WinGetVolume(win_struct *win);
extern void WinSetVolume(win_struct *win, const gfloat *vol);

extern void WinOpen(win_struct *win, const gchar *path);
extern void WinPlay(win_struct *win);   
extern void WinPause(win_struct *win);
extern void WinStop(win_struct *win);
extern void WinRewind(win_struct *win);

extern void WinManage(win_struct *win);

extern void WinDSPViewSetDisplayStyle(
	win_struct *win, dspview_display_style display_style
);
extern void WinDSPViewDisplayStyleNext(win_struct *win);
extern void WinDSPViewDisplayStylePrev(win_struct *win);

extern win_opid_struct *WinOPIDNew(
	win_opid id, win_struct *win
);
extern void WinOPIDDelete(win_opid_struct *opid);
extern GList *WinOPIDListNew(win_struct *win);
extern void WinOPIDListDelete(GList *glist);
extern win_opid_struct *WinOPIDListFind(GList *glist, win_opid id);

extern win_struct *WinNew(gpointer core_ptr);
extern void WinUpdate(win_struct *win);
extern void WinSetBusy(win_struct *win, gboolean busy);
extern gboolean WinToolBarIsShown(win_struct *win);
extern void WinToolBarSetShow(win_struct *win, gboolean show);
extern void WinToolBarUpdateItems(win_struct *win);
extern gboolean WinDisplayIsShown(win_struct *win);
extern void WinDisplaySetShow(win_struct *win, gboolean show);
extern gboolean WinVolumeIsShown(win_struct *win);
extern void WinVolumeSetShow(win_struct *win, gboolean show);
extern gboolean WinStatusBarIsShown(win_struct *win);
extern void WinStatusBarSetShow(win_struct *win, gboolean show);
extern void WinStatusProgress(
	win_struct *win, gfloat val, gboolean allow_gtk_iteration
);
extern void WinStatusMessage(
	win_struct *win, const gchar *mesg, gboolean allow_gtk_iteration
);
extern gboolean WinIsMapped(win_struct *win);
extern void WinMap(win_struct *win);
extern void WinUnmap(win_struct *win);
extern void WinDelete(win_struct *win);


#endif	/* WIN_H */
